/* #define DEBUGTRACE_ENABLED */
#include "debugtrace.hpp"
#include "lasp_config.h"

#if LASP_HAS_ULDAQ == 1
#include "lasp_daqconfig.h"
#include "lasp_uldaq.h"
#include <algorithm>
#include <atomic>
#include <cassert>
#include <chrono>
#include <gsl/gsl-lite.hpp>
#include <iostream>
#include <memory>
#include <mutex>
#include <stdexcept>
#include <thread>
#include <uldaq.h>
#include <vector>

using namespace std::literals::chrono_literals;
using std::atomic;
using std::cerr;
using std::endl;
using rte = std::runtime_error;

#include "debugtrace.hpp"
DEBUGTRACE_VARIABLES;

const us MAX_DEV_COUNT_PER_API = 100;
/**
 * @brief Reserve some space for an error message from UlDaq
 */
const us UL_ERR_MSG_LEN = 512;

/**
 * @brief Show the error to default error stream and return a string
 * corresponding to the error
 *
 * @param err Error string
 */
string showErr(UlError err) {
  string errstr;
  errstr.reserve(UL_ERR_MSG_LEN);
  if (err != ERR_NO_ERROR) {
    char errmsg[UL_ERR_MSG_LEN];
    errstr = "UlDaq API Error: ";
    ulGetErrMsg(err, errmsg);
    errstr += errmsg;

    std::cerr << "\b\n**************** UlDAQ backend error **********\n";
    std::cerr << errstr << std::endl;
    std::cerr << "***********************************************\n\n";
    return errstr;
  }
  return errstr;
}

class DT9837A : public Daq {

  DaqDeviceHandle _handle = 0;
  std::mutex _daqmutex;

  std::thread _thread;
  atomic<bool> _stopThread{false};
  atomic<StreamStatus> _streamStatus;

  const us _nFramesPerBlock;

  void threadFcn(InDaqCallback inCallback, OutDaqCallback outcallback);

public:
  DaqDeviceHandle getHandle() const { return _handle; }
  DT9837A(const DeviceInfo &devinfo, const DaqConfiguration &config);

  ~DT9837A() {
    UlError err;
    if (isRunning()) {
      stop();
    }

    if (_handle) {
      err = ulDisconnectDaqDevice(_handle);
      showErr(err);
      err = ulReleaseDaqDevice(_handle);
      showErr(err);
    }
  }

  bool isRunning() const {
    DEBUGTRACE_ENTER;
    return _thread.joinable();
  }
  virtual void start(InDaqCallback inCallback,
                     OutDaqCallback outCallback) override final;

  virtual StreamStatus getStreamStatus() const override {
    return _streamStatus;
  }

  void stop() override final {
    DEBUGTRACE_ENTER;
    StreamStatus status = _streamStatus;
    if (!isRunning()) {
      throw rte("No data acquisition running");
    }

    _stopThread = true;
    if (_thread.joinable()) {
      _thread.join();
    }
    _stopThread = false;
    status.isRunning = false;
    _streamStatus = status;
  }

  friend class InBufHandler;
  friend class OutBufHandler;
};

void DT9837A::start(InDaqCallback inCallback, OutDaqCallback outCallback) {
  DEBUGTRACE_ENTER;
  if (isRunning()) {
    throw rte("DAQ is already running");
  }
  if (neninchannels() > 0) {
    if (!inCallback)
      throw rte("DAQ requires a callback for input data");
  }
  if (nenoutchannels() > 0) {
    if (!outCallback)
      throw rte("DAQ requires a callback for output data");
  }
  assert(neninchannels() + nenoutchannels() > 0);
  _thread = std::thread(&DT9837A::threadFcn, this, inCallback, outCallback);
}

class BufHandler {
protected:
  DT9837A &daq;
  const DataTypeDescriptor dtype_descr;
  us nchannels, nFramesPerBlock;
  double samplerate;
  std::vector<double> buf;
  bool topenqueued, botenqueued;

  us increment = 0;

  us totalFramesCount = 0;
  long long buffer_mid_idx;

public:
  BufHandler(DT9837A &daq, const us nchannels)
      : daq(daq), dtype_descr(daq.dtypeDescr()), nchannels(nchannels),
        nFramesPerBlock(daq.framesPerBlock()), samplerate(daq.samplerate()),
        buf(2 * nchannels *
                nFramesPerBlock, // Watch the two here, the top and the bottom!
            0),
        buffer_mid_idx(nchannels * nFramesPerBlock) {
    assert(nchannels > 0);
  }
};
class InBufHandler : public BufHandler {
  bool monitorOutput;
  InDaqCallback cb;

public:
  InBufHandler(DT9837A &daq, InDaqCallback cb)
      : BufHandler(daq, daq.neninchannels()), cb(cb)

  {
    DEBUGTRACE_ENTER;
    assert(daq.getHandle() != 0);

    monitorOutput = daq.monitorOutput;

    DaqInScanFlag inscanflags = DAQINSCAN_FF_DEFAULT;
    ScanOption scanoptions = SO_CONTINUOUS;
    UlError err = ERR_NO_ERROR;

    std::vector<DaqInChanDescriptor> indescs;
    boolvec eninchannels_without_mon = daq.eninchannels(false);

    // Initialize input, if any
    dvec ranges = daq.inputRangeForEnabledChannels(false);
    for (us chin = 0; chin < 4; chin++) {
      if (eninchannels_without_mon[chin] == true) {
        DaqInChanDescriptor indesc;
        indesc.type = DAQI_ANALOG_SE;
        indesc.channel = chin;

        double rangeval = ranges.at(chin);
        Range rangenum;
        if (fabs(rangeval - 1.0) < 1e-8) {
          rangenum = BIP1VOLTS;
        } else if (fabs(rangeval - 10.0) < 1e-8) {
          rangenum = BIP10VOLTS;
        } else {
          std::cerr << "Fatal: input range value is invalid" << endl;
          return;
        }
        indesc.range = rangenum;
        indescs.push_back(indesc);
      }
    }

    // Add possibly last channel as monitor
    if (monitorOutput) {
      DaqInChanDescriptor indesc;
      indesc.type = DAQI_DAC;
      indesc.channel = 0;
      /// The output only has a range of 10V, therefore the monitor of the
      /// output also has to be set to this value.
      indesc.range = BIP10VOLTS;
      indescs.push_back(indesc);
    }
    assert(indescs.size() == nchannels);

    DEBUGTRACE_MESSAGE("Starting input scan");

    err = ulDaqInScan(daq.getHandle(), indescs.data(), nchannels,
                      2 * nFramesPerBlock, // Watch the 2 here!
                      &samplerate, scanoptions, inscanflags, buf.data());
    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Could not start input DAQ");
    }
  }
  void start() {

    ScanStatus status;
    TransferStatus transferStatus;
    UlError err = ulDaqInScanStatus(daq.getHandle(), &status, &transferStatus);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Unable to start input on DAQ");
    }

    totalFramesCount = transferStatus.currentTotalCount;
    topenqueued = true;
    botenqueued = true;
  }

  /**
   * @brief InBufHandler::operator()()
   *
   * @return true on success
   */
  bool operator()() {

    /* DEBUGTRACE_ENTER; */

    bool ret = true;

    auto runCallback = ([&](us totalOffset) {
      /* DEBUGTRACE_ENTER; */

      DaqData data(nFramesPerBlock, nchannels,
                   DataTypeDescriptor::DataType::dtype_fl64);

      us monitorOffset = monitorOutput ? 1 : 0;
      /* /// Put the output monitor in front */
      if (monitorOutput) {
        for (us frame = 0; frame < nFramesPerBlock; frame++) {
          data.value<double>(frame, 0) =
              buf[totalOffset + (frame * nchannels) + (nchannels - 1)];
        }
      }

      for (us channel = 0; channel < nchannels - monitorOffset; channel++) {
        /* DEBUGTRACE_PRINT(channel); */
        for (us frame = 0; frame < nFramesPerBlock; frame++) {
          data.value<double>(frame, channel + monitorOffset) =
              buf[totalOffset + (frame * nchannels) + channel];
        }
      }
      return cb(data);
    });

    ScanStatus status;
    TransferStatus transferStatus;

    UlError err = ulDaqInScanStatus(daq.getHandle(), &status, &transferStatus);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      return false;
    }

    increment = transferStatus.currentTotalCount - totalFramesCount;
    totalFramesCount += increment;

    if (increment > nFramesPerBlock) {
      cerr << "Error: overrun for input of DAQ!" << endl;
      return false;
    }
    assert(status == SS_RUNNING);

    if (transferStatus.currentIndex < (long long)buffer_mid_idx) {
      topenqueued = false;
      if (!botenqueued) {
        ret = runCallback(nchannels * nFramesPerBlock);
        botenqueued = true;
      }
    } else {
      botenqueued = false;
      if (!topenqueued) {
        ret = runCallback(0);
        topenqueued = true;
      }
    }
    return ret;
  }
  ~InBufHandler() {
    // At exit of the function, stop scanning.
    DEBUGTRACE_ENTER;
    UlError err = ulDaqInScanStop(daq.getHandle());
    if (err != ERR_NO_ERROR) {
      showErr(err);
    }
  }
};

class OutBufHandler : public BufHandler {
  OutDaqCallback cb;

public:
  OutBufHandler(DT9837A &daq, OutDaqCallback cb)
      : BufHandler(daq, daq.nenoutchannels()), cb(cb) {

    DEBUGTRACE_MESSAGE("Starting output scan");
    DEBUGTRACE_PRINT(nchannels);
    AOutScanFlag outscanflags = AOUTSCAN_FF_DEFAULT;
    ScanOption scanoptions = SO_CONTINUOUS;
    UlError err =
        ulAOutScan(daq.getHandle(), 0, 0, BIP10VOLTS,
                   2 * nFramesPerBlock, // Watch the 2 here!
                   &samplerate, scanoptions, outscanflags, buf.data());

    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Unable to start output on DAQ");
    }
  }
  void start() {

    ScanStatus status;
    TransferStatus transferStatus;

    UlError err = ulAOutScanStatus(daq.getHandle(), &status, &transferStatus);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Unable to start output on DAQ");
    }
    if (status != SS_RUNNING) {
      throw rte("Unable to start output on DAQ");
    }
    totalFramesCount = transferStatus.currentTotalCount;
    topenqueued = true;
    botenqueued = true;
  }
  /**
   * @brief OutBufHandler::operator()
   *
   * @return true on success
   */
  bool operator()() {

    /* DEBUGTRACE_ENTER; */
    bool res = true;
    assert(daq.getHandle() != 0);

    UlError err = ERR_NO_ERROR;

    ScanStatus status;
    TransferStatus transferStatus;

    err = ulAOutScanStatus(daq.getHandle(), &status, &transferStatus);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      return false;
    }
    if (status != SS_RUNNING) {
      return false;
    }
    increment = transferStatus.currentTotalCount - totalFramesCount;
    totalFramesCount += increment;

    if (increment > nFramesPerBlock) {
      cerr << "Error: underrun for output of DAQ!" << endl;
      return false;
    }

    if (transferStatus.currentIndex < buffer_mid_idx) {
      topenqueued = false;
      if (!botenqueued) {
        DaqData d(nFramesPerBlock,1,
                               DataTypeDescriptor::DataType::dtype_fl64);
        res = cb(d);
        d.copyToRaw(0, reinterpret_cast<byte_t*>(&(buf[buffer_mid_idx])));

        botenqueued = true;
      }
    } else {
      botenqueued = false;
      if (!topenqueued) {
        DaqData d(nFramesPerBlock,1,
                               DataTypeDescriptor::DataType::dtype_fl64);
        res = cb(d);
        d.copyToRaw(0, reinterpret_cast<byte_t*>(&(buf[0])));

        topenqueued = true;
      }
    }
    return res;
  }

  ~OutBufHandler() {
    DEBUGTRACE_ENTER;
    UlError err = ulAOutScanStop(daq.getHandle());
    if (err != ERR_NO_ERROR) {
      showErr(err);
    }
  }
};

void DT9837A::threadFcn(InDaqCallback inCallback, OutDaqCallback outCallback) {

  DEBUGTRACE_ENTER;

  /* cerr << "******************\n" */
  /*   "Todo: the current way of handling timing in this DAQ thread is not " */
  /*   "really robust, due " */
  /*   "to input / output callbacks that can be too time-consuming. We have " */
  /*   "to fix the " */
  /*   "sleep_for to properly deal with longer callbacks." */
  /*   "\n*****************" */
  /*   << endl; */
  try {

    std::unique_ptr<OutBufHandler> obh;
    std::unique_ptr<InBufHandler> ibh;

    StreamStatus status = _streamStatus;
    status.isRunning = true;
    _streamStatus = status;

    if (nenoutchannels() > 0) {
      assert(outCallback);
      obh = std::make_unique<OutBufHandler>(*this, outCallback);
    }
    if (neninchannels() > 0) {
      assert(inCallback);
      ibh = std::make_unique<InBufHandler>(*this, inCallback);
    }
    if (obh)
      obh->start();
    if (ibh)
      ibh->start();

    const double sleeptime_s =
        static_cast<double>(_nFramesPerBlock) / (16 * samplerate());
    const us sleeptime_us = static_cast<us>(sleeptime_s * 1e6);

    while (!_stopThread) {
      if (ibh) {
        if (!(*ibh)()) {
          _stopThread = true;
        }
      }
      if (obh) {
        if (!(*obh)()) {
          _stopThread = true;
        }
      } else {
        std::this_thread::sleep_for(std::chrono::microseconds(sleeptime_us));
      }
    }
  } catch (rte &e) {

    StreamStatus status = _streamStatus;
    status.isRunning = false;
    status.errorType = StreamStatus::StreamError::systemError;
    _streamStatus = status;

    cerr << "\n******************\n";
    cerr << "Catched error in UlDAQ thread: " << e.what() << endl;
    cerr << "\n******************\n";
  }
  StreamStatus status = _streamStatus;

  status.isRunning = false;
  _streamStatus = status;
  _stopThread = false;
}

std::unique_ptr<Daq> createUlDaqDevice(const DeviceInfo &devinfo,
                                       const DaqConfiguration &config) {
  return std::make_unique<DT9837A>(devinfo, config);
}

DT9837A::DT9837A(const DeviceInfo &devinfo, const DaqConfiguration &config)
    : Daq(devinfo, config),
      _nFramesPerBlock(availableFramesPerBlock.at(framesPerBlockIndex)) {

  // Some sanity checks
  if (inchannel_config.size() != 4) {
    throw rte("Invalid length of enabled inChannels vector");
  }

  if (outchannel_config.size() != 1) {
    throw rte("Invalid length of enabled outChannels vector");
  }

  if (_nFramesPerBlock < 24 || _nFramesPerBlock > 8192) {
    throw rte("Unsensible number of samples per block chosen");
  }

  if (samplerate() < 10000 || samplerate() > 51000) {
    throw rte("Invalid sample rate");
  }

  DaqDeviceDescriptor devdescriptors[MAX_DEV_COUNT_PER_API];
  DaqDeviceDescriptor descriptor;
  DaqDeviceInterface interfaceType = ANY_IFC;

  UlError err;

  us numdevs = MAX_DEV_COUNT_PER_API;
  err = ulGetDaqDeviceInventory(interfaceType, devdescriptors,
                                (unsigned *)&numdevs);
  if (err != ERR_NO_ERROR) {
    throw rte("Device inventarization failed");
  }

  if ((us)api_specific_devindex >= numdevs) {
    throw rte("Device number {deviceno} too high {err}. This could "
              "happen when the device is currently not connected");
  }

  descriptor = devdescriptors[api_specific_devindex];

  // get a handle to the DAQ device associated with the first descriptor
  _handle = ulCreateDaqDevice(descriptor);

  if (_handle == 0) {
    throw rte("Unable to create a handle to the specified DAQ "
              "device. Is the device currently in use? Please make sure to set "
              "the DAQ configuration in duplex mode if simultaneous input and "
              "output is required.");
  }

  err = ulConnectDaqDevice(_handle);
  if (err != ERR_NO_ERROR) {
    ulReleaseDaqDevice(_handle);
    _handle = 0;
    throw rte(string("Unable to connect to device: " + showErr(err)));
  }

  for (us ch = 0; ch < 4; ch++) {

    err = ulAISetConfigDbl(_handle, AI_CFG_CHAN_SENSOR_SENSITIVITY, ch, 1.0);
    showErr(err);
    if (err != ERR_NO_ERROR) {
      throw rte("Fatal: could normalize channel sensitivity");
    }

    CouplingMode cm = inchannel_config.at(ch).ACCouplingMode ? CM_AC : CM_DC;
    err = ulAISetConfig(_handle, AI_CFG_CHAN_COUPLING_MODE, ch, cm);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Fatal: could not set AC/DC coupling mode");
    }

    IepeMode iepe =
        inchannel_config.at(ch).IEPEEnabled ? IEPE_ENABLED : IEPE_DISABLED;
    err = ulAISetConfig(_handle, AI_CFG_CHAN_IEPE_MODE, ch, iepe);
    if (err != ERR_NO_ERROR) {
      showErr(err);
      throw rte("Fatal: could not set IEPE mode");
    }
  }
}
void fillUlDaqDeviceInfo(std::vector<DeviceInfo> &devinfolist) {

  DEBUGTRACE_ENTER;

  UlError err;
  unsigned int numdevs = MAX_DEV_COUNT_PER_API;

  DaqDeviceDescriptor devdescriptors[MAX_DEV_COUNT_PER_API];
  DaqDeviceDescriptor descriptor;
  DaqDeviceInterface interfaceType = ANY_IFC;

  err = ulGetDaqDeviceInventory(interfaceType, devdescriptors,
                                static_cast<unsigned *>(&numdevs));

  if (err != ERR_NO_ERROR) {
    throw rte("UlDaq device inventarization failed");
  }

  for (unsigned i = 0; i < numdevs; i++) {

    descriptor = devdescriptors[i];

    DeviceInfo devinfo;
    devinfo.api = uldaqapi;
    string name, interface;
    if (string(descriptor.productName) != "DT9837A") {
      throw rte("Unknown UlDAQ type");
    }

    switch (descriptor.devInterface) {
    case USB_IFC:
      name = "USB - ";
      break;
    case BLUETOOTH_IFC:
      /* devinfo. */
      name = "Bluetooth - ";
      break;

    case ETHERNET_IFC:
      /* devinfo. */
      name = "Ethernet - ";
      break;
    default:
      name = "Uknown interface = ";
    }

    name += string(descriptor.productName) + " " + string(descriptor.uniqueId);
    devinfo.device_name = std::move(name);

    devinfo.api_specific_devindex = i;
    devinfo.availableDataTypes.push_back(
        DataTypeDescriptor::DataType::dtype_fl64);
    devinfo.prefDataTypeIndex = 0;

    devinfo.availableSampleRates = {8000,  10000, 11025, 16000, 20000,
                                    22050, 24000, 32000, 44056, 44100,
                                    47250, 48000, 50000, 50400, 51000};

    devinfo.prefSampleRateIndex = 11;

    devinfo.availableFramesPerBlock = {512, 1024, 2048, 4096, 8192};

    devinfo.availableInputRanges = {1.0, 10.0};
    devinfo.prefInputRangeIndex = 0;

    devinfo.ninchannels = 4;
    devinfo.noutchannels = 1;

    devinfo.hasInputIEPE = true;
    devinfo.hasInputACCouplingSwitch = true;
    devinfo.hasInputTrigger = true;

    devinfo.hasInternalOutputMonitor = true;

    // Finally, this devinfo is pushed back in list
    devinfolist.push_back(devinfo);
  }
}
#endif // LASP_HAS_ULDAQ
